import java.util.*; // gives access to class Arrays

/** An instance is a movie and, if known, ratings from a fixed set of critics.*/
public class Movie implements Comparable, MetricEmbedded {
    
    private String name; // name of the movie
    
    /** ratings[i]= either critic i's positive rating of the movie
      * or NOT_RATED=0 if the critic i didn't rate it,
      * where it is assumed that each critic has a unique ID number
      * in 0..NUMCRITICS-1; */
    private static final int NUMCRITICS= 3; // #@ FOR TESTING
    private static final int NOT_RATED= 0;
    private int[] ratings= new int[NUMCRITICS]; // init: all = NOT_RATED
    
    /** Constructor: a movie with this name */
    public Movie(String name) {
        this.name= name;
    }
    
    /** Constructor: a movie with this name and with critic ratings r.
      * Throws IllegalArgumentException if r.length!=NUMCRITICS */
    public Movie(String name, int[] r) {
        this(name);
        
        if (r.length != NUMCRITICS) {
            throw new IllegalArgumentException("length of ratings array r for movie " 
                                                   + name + " doesn't match number" 
                                                   + " of grouplens critics, " 
                                                   + NUMCRITICS);
        }
        
        ratings= r;
    }
        
    
    /** = -1, 0, or 1 depending on alphabetical precedence/equality/postedence of
      * this Movie's name and c's name.
      * Throws a ClassCastException if c isn't a Movie*/
    public int compareTo(Object c) {
        if (!(c  instanceof  Movie)) 
            throw new ClassCastException("argument is not a Movie"); 
        
        Movie comparee= (Movie) c;
        
        // use String's compareTo, as run on the Movies' names
        return this.name.compareTo(comparee.name); 
        
    }
    
    /** = name of this Movie.  (The ratings array is generally too long
      * to be useful to see).*/
    public String toString() {
        return "\"" + name  + "\"";
    }
    
    
    /** Show how sorting using the Comparable interface works */
    public static void sortWithCompareToDemo() {
        Movie starwars= new Movie("Star Wars");
        Movie casablanca= new Movie("Casablanca");      
        Movie love= new Movie("Must Love Dogs");      
        Movie sc= new Movie("Sex and the City");
        Movie matrix= new Movie("Matrix, The");
        Movie lep= new Movie("Leprechaun 2");
        
        Movie[] arr= new Movie[] {starwars, casablanca, love, sc, matrix, lep};
        
 
        // Arrays.toString is in Java 5: handles pretty-printing of arrays
        System.out.println("Array arr before sorting: " + Arrays.toString(arr));
     
        // Calling Arrays.sort rather than our own sorting routine!
        Arrays.sort(arr); 
        System.out.println("Array arr after Arrays.sort(arr): " + Arrays.toString(arr));
        
        // reverse the order
        Arrays.sort(arr, Collections.reverseOrder());
        System.out.println("Array arr after Arrays.sort(arr, Collections.reverseOrder()): " 
                               + Arrays.toString(arr));
    }
        
    /** Return the distance of obj from this Movie,
      * as measured in Euclidean distance on ratings vectors.
      * Returns a ClassCastException if obj isn't a Movie. */
    public double distanceTo(Object obj) {
        if (!(obj instanceof Movie)) {
            throw new ClassCastException("argument is not a Movie"); 
        }
        Movie m= (Movie) obj;
        
        if (ratings.length != NUMCRITICS || m.ratings.length != NUMCRITICS) {
            throw new IllegalArgumentException("the number of known critics doesn't" +
                                               "match the number of ratings for at least one of " 
                                                   + toString() + " or " + m.toString());
        }
        
        if (this.equals(obj)) {
            return 0.0;
        }
        
        double distSq= 0.0;
        /** inv: Euclidean distance: sqrt of sum for i=0..NUMCRITICS-1
          * of (ratings[i] - m.ratings[i])^2, except that we ignore
          * those i for which critic i did not rate at least one of
          * the two movies being compared.
          * Basically, every difference some critic's rating of the 
          * movie adds to the distance.
          * (We are treating the ratings as mathematical vectors in 
          * a high-dimensional space, in general).
           */
        for (int i=0; i < NUMCRITICS; i= i+1) {
            if (!(ratings[i] == NOT_RATED || m.ratings[i] == NOT_RATED)) {
                double diffi= ratings[i] - m.ratings[i];
                distSq= distSq + diffi*diffi;
            }
        }
        
        return Math.sqrt(distSq);
    }
    
     /** Show how computing distance between ratings 
       * (mathematical) vectors works */
    public static void sortWithDistComparatorDemo() {
        Movie starwars= new Movie("Star Wars", new int[] {5, 5, 4});
        Movie casablanca= new Movie("Casablanca", new int[] {4, 5, 5});      
        Movie love= new Movie("Must Love Dogs", new int[] {2, 3, 5});      
        Movie sc= new Movie("Sex and the City", new int[] {3, 3, 5});
        Movie matrix= new Movie("Matrix, The", new int[] {5, 3, 2});
        Movie lep= new Movie("Leprechaun 2", new int[] {2, 1, 1});
        
        Movie[] arr= new Movie[] {starwars, casablanca, love, sc, matrix, lep};
        
        Movie fixedMovie= love;
        
        // inv: distance between fixedMovie and arr[0..i-1] has been printed
        for (int i=0; i < arr.length; i= i+1) {
            System.out.println("distance between " + fixedMovie + " and " + arr[i] + ": "  
                               + fixedMovie.distanceTo(arr[i]));
        }
        
        // show how array sorting works with an object-dependent comparator
        System.out.println("\nArray before sorting: " + Arrays.toString(arr));
        MetricEmbeddings.sortByDistanceTo(arr, fixedMovie); 
        System.out.println("sorted arr using distance to " + fixedMovie.toString() +
                           " as compare for Comparator: "
                               + Arrays.toString(arr));
    }
        
    /** = -1, 0, or 1 as obj1 is a Movie with distance to this Movie less than,
      * equal to, or greater than Movie obj2.
      * Throws a ClassCastException if either obj1 or obj2 are not Movies. */
    public int compare(Object obj1, Object obj2) {
        if (!(obj1 instanceof Movie && obj2 instanceof Movie)) {
            throw new ClassCastException("compare: at least one of these is" +
                                         "not a Movie: " + obj1.toString() +
                                         ", " + obj2.toString());
        }
        if (this.distanceTo(obj1) - this.distanceTo(obj2) < 0) {
            return -1;
        }
        else if (this.distanceTo(obj1) - this.distanceTo(obj2) > 0) {
            return 1;
        }
        else {
            return 0; 
        }
        
        // #@ thought question: why can't you just do
        // "return (int)(this.distanceTo(obj1) - this.distanceTo(obj2))"?
    }
    
}
